#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "fat32.h"

#define IS_EOC(id)     ((id) >= 0x0FFFFFF8)
#define DIR_ENTRY_SIZE (32)
#define MIN(a, b)      ((a) < (b) ? (a) : (b))


/** pomocna funkce, ktera precte zadane sektor(y) do bufferu */
static void sector_read(struct fat32 *fs, uint32_t sector, uint32_t count, uint8_t *buf)
{
	fseek(fs->image, (long) sector * fs->BPB_BytsPerSec, SEEK_SET);
	fread(buf, fs->BPB_BytsPerSec, count, fs->image);
}

/** pomocna funkce, ktera nacte cluster s danym ID do bufferu */
static void cluster_read(struct fat32 *fs, uint32_t clusterId, uint8_t *buf)
{
	uint32_t sector = ((clusterId - 2) * fs->BPB_SecPerClus) + fs->FirstDataSector;
	sector_read(fs, sector, fs->BPB_SecPerClus, buf);
}

/** otevre soubor, ktery je identifikovany prvnim svym clusterem */
struct fat32_fd *fat32_open(struct fat32 *fs, uint32_t clusterId, int32_t fileSize)
{
	struct fat32_fd *fd = malloc(sizeof(struct fat32_fd) + fs->ClusterSz);
	fd->fs              = fs;
	fd->currentCluster  = clusterId;
	fd->isDirectory     = (fileSize == FAT32_DIR);
	fd->fileSize        = fileSize;
	fd->fileOffset      = 0;
	fd->bufOffset       = 0;
	fd->bufSize         = 0;

	// prazdny soubor
	if (IS_EOC(clusterId)) return fd;

	cluster_read(fs, clusterId, fd->buf);
	if (fd->isDirectory) {
		fd->bufSize = fs->ClusterSz;
	} else {
		fd->bufSize = MIN(fs->ClusterSz, fileSize);
	}
	return fd;
}

/** ukonci praci s otevrenym souborem */
void fat32_close(struct fat32_fd *fd)
{
	free(fd);
}

/** pro aktualni cluster, vraci cluster nasledujici */
static uint32_t fat32_next_cluster_id(struct fat32 *fs, uint32_t currentCluster)
{
	uint8_t  buf[fs->BPB_BytsPerSec];
	uint32_t fat_offset = currentCluster * 4;
	uint32_t fat_sector = fs->BPB_RsvdSecCnt + (fat_offset / fs->BPB_BytsPerSec);
	uint32_t fat_entry  = fat_offset % fs->BPB_BytsPerSec;

	sector_read(fs, fat_sector, 1, buf);
	
	uint32_t rawId = *((uint32_t *) (buf + fat_entry));
	return rawId & 0x0fffffff;
}

/** pomocna funkce, ktera nacte do bufferu nasledujici cluster */
static int fat32_next_cluster(struct fat32_fd *fd)
{
	uint32_t next_cluster = fat32_next_cluster_id(fd->fs, fd->currentCluster);
	if (IS_EOC(next_cluster)) return 0;

	cluster_read(fd->fs, next_cluster, fd->buf);
	fd->bufOffset = 0;
	fd->currentCluster = next_cluster;

	return 1;
}


/**
 * cte ze souboru `size' bytu a uklada je do bufferu
 * vraci pocet skutecne prectenych bytu dat
 */ 
size_t fat32_read(struct fat32_fd *fd, size_t size, uint8_t *dest_buf)
{
	size_t total_read = 0;
	while (1) {
		uint32_t available  = fd->bufSize - fd->bufOffset;
		uint32_t read_bytes = MIN(size, available);
		
		memcpy(dest_buf, fd->buf, read_bytes);
		dest_buf       += read_bytes;
		size           -= read_bytes;
		total_read     += read_bytes;
		fd->bufOffset  += read_bytes;
		fd->fileOffset += read_bytes;
		if (size == 0) return total_read;
		
		int success = fat32_next_cluster(fd);
		if (!success) return total_read;

		// posledni cluster souboru
		if ((fd->fileSize - fd->fileOffset) <= fd->fs->ClusterSz) {
			fd->bufSize = fd->fileSize - fd->fileOffset;
		}
		
	}
}


/**
 * Ze souboru, ktery je adresarem, precte jeden (platny) zaznam
 * a ulozi jej do parametru entry, posune se na dalsi zaznam
 * a vrati kladnou hodnotu. Pokud neexistuje dalsi validni zaznam,
 * vraci 0.
 */
int fat32_read_dir(struct fat32_fd *fd, struct fat32_dir_entry *entry)
{
	while (1) {
		if (fd->bufOffset >= fd->fs->ClusterSz) {
			int has_next = fat32_next_cluster(fd);
			if (!has_next) return 0;
		}


		uint8_t entry_data[DIR_ENTRY_SIZE];
		memcpy(entry_data, fd->buf + fd->bufOffset, DIR_ENTRY_SIZE);
		fd->bufOffset += DIR_ENTRY_SIZE;

		/* invalid record */
		if (entry_data[0] == 0xe5) continue;
		if (entry_data[0] == 0x00) return 0;
		if (entry_data[0] < 0x20) continue;

		memcpy(&entry->fileName, entry_data + 0, 11);
		memcpy(&entry->attr, entry_data + 11, 1);
		memcpy(&entry->fstClusterHI, entry_data + 20, 2);
		memcpy(&entry->fstClusterLO, entry_data + 26, 2);
		memcpy(&entry->fileSize,     entry_data + 28, 4);
		entry->fstClusterId = ((uint32_t) entry->fstClusterHI) << 16 | entry->fstClusterLO;

		if (entry->attr != ATTR_LONG_NAME) return 1;
	}
}

/** pripravi strukturu, pro praci se souborovym systemem */
void fat32_init(struct fat32 *fs, FILE *image)
{
	uint8_t sector[512];
	fread(sector, 1, 512, image);

	fs->image = image;
	memcpy(&fs->BPB_BytsPerSec, sector + 11, 2);
	memcpy(&fs->BPB_SecPerClus, sector + 13, 1);
	memcpy(&fs->BPB_RsvdSecCnt, sector + 14, 2);
	memcpy(&fs->BPB_NumFATs,    sector + 16, 1);
	memcpy(&fs->BPB_TotSec32,   sector + 32, 4);
	memcpy(&fs->BPB_FATSz32,    sector + 36, 4);
	memcpy(&fs->BPB_ExtFlags,   sector + 40, 2);
	memcpy(&fs->BPB_FSVer,      sector + 42, 2);
	memcpy(&fs->BPB_RootClus,   sector + 44, 4);
	memcpy(&fs->BPB_FSInfo,     sector + 48, 2);
	memcpy(&fs->BS_BootSig,     sector + 66, 1);
	memcpy(&fs->BS_VolID,       sector + 67, 4);
	memcpy(&fs->BS_VolLab,      sector + 71, 11);
	memcpy(&fs->BS_FilSysType,  sector + 82, 8);

	fs->FirstDataSector = fs->BPB_RsvdSecCnt + (fs->BPB_NumFATs * fs->BPB_FATSz32);
	fs->DataSec = fs->BPB_TotSec32 - fs->BPB_RsvdSecCnt - (fs->BPB_NumFATs * fs->BPB_FATSz32); 
	fs->CountOfClusters = fs->DataSec / fs->BPB_SecPerClus;
	fs->ClusterSz = fs->BPB_SecPerClus * fs->BPB_BytsPerSec;
}
